## pure2nix



<!--
```graphviz

digraph {

label="pure2nix concept"

lockfile [style=filled,color=pink]
manifest [style=filled,color=lightblue]
sources [style=filled,color=lightblue]
checked_sources [style=filled,color=lightblue]
cache  [style=filled,color=pink]

importer [shape=Mrect,style=filled,color=pink]
consumer [shape=Mrect,style=filled,color=pink]
fetcher [shape=Mrect,style=filled,color=lightblue]

lockfile -> importer
lockfile -> consumer
importer -> manifest
manifest -> fetcher
fetcher -> sources
manifest -> checked_sources
sources -> checked_sources
checked_sources -> consumer
consumer -> cache

cache -> foo
foo[label="..."]
}
```
-->

| lockfile format          | importer           | consumer               | recursion | native parser | comments                                |
|--------------------------|--------------------|------------------------|-----------|---------------|-----------------------------------------|
| Cargo.lock               | :white_check_mark: | :white_check_mark:     | ?         | yes           | -                                       |
| npm package-lock.json v1 | Planned            | Planned                | for dev deps of git deps | yes           | -                                       |
| npm package-lock.json v2 | :white_check_mark: | :white_check_mark:     | for dev deps of git deps | yes           | -                                       |
| yarn.lock v1             | :white_check_mark: | WIP                    | ?         | yes           | -                                       |
| yarn berry lock          | Planned            | Planned                | ?         | yes           | -                                       |
| pnpm-lock.yaml           | Planned            | Planned                | ?         | yes           | -                                       |
| elixir / mix.lock        | Maybe              | Maybe                  | ?         | no            | requires custom fetcher to match hashes |
| poetry.lock              | Maybe              | Maybe                  | ?         | ?             | -                                       |
| git submodules           | Maybe              | Maybe                  | yes       | ?             | -                                       |
| gclient DEPS             | Maybe              | Maybe                  | yes       | no            | -                                       |


The importer parses a language package manager lockfile and converts it to a manifest.

The manifest is a json file describing what should be downloaded from where.

The consumer takes the downloaded files and converts them into a usable format. This step would for example include re-packing git repositories or extracting tarballs to match what the language package manager expects.

fetchers:
- individual eval-time downloads using builtins.fetchGit, builtins.fetchTree
- FOD fetcher: fetch all sources in a single big FOD derivation
- in the future: (real) recursive Nix and git-hashed trees

### native mode

```mermaid
graph TD


subgraph Eval-Time Emitted Paths
lockfile((Lockfile))
importer
manifest((Manifest))
fetcher
source1((Source 1))
source2((Source 2))
end

subgraph Fixed-Output Derivations
source3((Source 3))
source4((Source 4))
end

subgraph Fixed-Input Derivations
sources((Sources))
consumer
cache((Cache))
foo((...))
end


lockfile --> importer --> manifest
manifest --> fetcher
fetcher --> source3 --> sources
fetcher --> source4 --> sources
fetcher --> source1 --> sources
fetcher --> source2 --> sources

lockfile --> consumer

sources --> consumer --> cache --> foo

style manifest fill:lightblue,color:black
style fetcher fill:lightblue,color:black
style source1 fill:lightblue,color:black
style source2 fill:lightblue,color:black
style source3 fill:lightblue,color:black
style source4 fill:lightblue,color:black
style sources fill:lightblue,color:black

style lockfile fill:pink,color:black
style importer fill:pink,color:black
style consumer fill:pink,color:black
style cache fill:pink,color:black
```

We do not require an FOD hash. We rely on the individual hashes from the lockfile and the builtins.fetchGit primop to obtain git sources. Each source is downloaded individually.

This is guaranteed to be Import-From-Derivation free, as long as the lockfile is not from a Derivation. This ensures a cross architecture deploy can safely be split up into evaluation and build phases e.g. `colmena apply --build-on-target` will not error with `a 'x86_64-linux' with features {} is required to build '...yarn.nix', but I am a 'aarch64-linux' with features {...}`.


### FOD (nixpkgs compatible) mode

legend:  
ellipse: derivation  
rectangle: function  
blue: generalized  
pink: language-specific  

```mermaid
graph TD

subgraph Fixed-Output Derivations with Internet Access
lockfile((Lockfile))
fetcher

sources((Sources))
end

subgraph Fixed-Input Derivations

importer
manifest((Manifest))
checked_sources((Checked Sources))
consumer
cache((Cache))
foo((...))
end

lockfile --> importer --> manifest
manifest --> checked_sources
manifest --> fetcher --> sources --> checked_sources
checked_sources --> consumer --> cache --> foo
lockfile --> consumer

style manifest fill:lightblue,color:black
style fetcher fill:lightblue,color:black
style sources fill:lightblue,color:black
style checked_sources fill:lightblue,color:black

style lockfile fill:pink,color:black
style importer fill:pink,color:black
style consumer fill:pink,color:black
style cache fill:pink,color:black

```

In this mode the importer is run in a derivation, by running `nix eval` within the derivation. The 'outer' Nix has no knowledge of the package language manager dependencies.

The mechanism of the fetcher can vary, besides it being run in a single Fixed-Output Derivation to fetch all sources, producing a tree of all the dependency tarballs / git checkouts.  
The 'checked sources' derivation then cross-checks that the manifest used to build the sources is identical with the one that we want, throwing an error if the FOD-hash is out of date.

### Native mode with recursion

Recursion on FOD fetchers would lead to IFD, so we might have to look into using builtins instead of pkgs.fetchurl.

```mermaid
graph TD



subgraph Fixed-Output Derivations
source3((Source 3))
source4((Source 4))
source9((Source 9))
source10((Source 10))
end

subgraph Fixed-Input Derivations
sources((Sources))
sources2((Sources 2))
sources3((Sources 3))
consumer
cache((Cache))
foo((...))
end

subgraph Eval-Time Emitted Paths
lockfile((Lockfile))
importer
manifest((Manifest))
fetcher
source1((Source 1))
source2((Source 2))

source5((Source 5))
source6((Source 6))
source1 --> importer2[importer] --> manifest2((Manifest)) --> fetcher2[fetcher]
fetcher2 --> source5 --> sources2
fetcher2 --> source6 --> sources2
sources2 --> sources
source7((Source 7))
source8((Source 8))
source2 --> importer3[importer] --> manifest3((Manifest)) --> fetcher3[fetcher]
fetcher3 --> source7 --> sources3
fetcher3 --> source8 --> sources3
fetcher3 --> source9 --> sources3
fetcher3 --> source10 --> sources3
sources3 --> sources
end

lockfile --> importer --> manifest
manifest --> fetcher
fetcher --> source1 --> sources
fetcher --> source2 --> sources
fetcher --> source3 --> sources
fetcher --> source4 --> sources

lockfile --> consumer

sources --> consumer --> cache --> foo

style manifest fill:lightblue,color:black
style manifest2 fill:lightblue,color:black
style manifest3 fill:lightblue,color:black
style fetcher fill:lightblue,color:black
style fetcher2 fill:lightblue,color:black
style fetcher3 fill:lightblue,color:black
style source1 fill:lightblue,color:black
style source2 fill:lightblue,color:black
style source3 fill:lightblue,color:black
style source4 fill:lightblue,color:black
style source5 fill:lightblue,color:black
style source6 fill:lightblue,color:black
style source7 fill:lightblue,color:black
style source8 fill:lightblue,color:black
style source9 fill:lightblue,color:black
style source10 fill:lightblue,color:black
style sources fill:lightblue,color:black
style sources2 fill:lightblue,color:black
style sources3 fill:lightblue,color:black

style lockfile fill:pink,color:black
style importer fill:pink,color:black
style importer2 fill:pink,color:black
style importer3 fill:pink,color:black
style consumer fill:pink,color:black
style cache fill:pink,color:black
```

### FOD mode with recursion

Only the top level manifest is available without relying Fixod-Output derivation, so we would rely on the importer behavior not changing for the sub-graphs

```mermaid
graph TD

subgraph Fixed-Output Derivations with Internet Access
lockfile((Lockfile))

subgraph Sources FOD
fetcher
sources((Sources))

fetcher --> source1((Source 1)) --> sources
fetcher --> source2((Source 2)) --> sources

source2 --> importer2[importer] --> manifest2((Manifest)) --> fetcher2[fetcher]
fetcher2 --> source3((Source 3)) --> sources
fetcher2 --> source4((Source 4)) --> sources

end
end

subgraph Fixed-Input Derivations

importer
manifest((Manifest))
checked_sources((Checked Sources))
consumer
cache((Cache))
foo((...))
end

lockfile --> importer --> manifest
manifest --> checked_sources
manifest --> fetcher --> sources --> checked_sources
checked_sources --> consumer --> cache --> foo
lockfile --> consumer

style manifest fill:lightblue,color:black
style manifest2 fill:lightblue,color:black
style fetcher fill:lightblue,color:black
style fetcher2 fill:lightblue,color:black
style source1 fill:lightblue,color:black
style source2 fill:lightblue,color:black
style source3 fill:lightblue,color:black
style source4 fill:lightblue,color:black
style sources fill:lightblue,color:black
style checked_sources fill:lightblue,color:black

style lockfile fill:pink,color:black
style importer fill:pink,color:black
style importer2 fill:pink,color:black
style consumer fill:pink,color:black
style cache fill:pink,color:black
```

